// C-Meter
//
//    Copyright (C) 2017  Michal Kopera
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

//programing firmware (with usbasp programmer):
//avrdude -p m48p -c usbasp -U flash:w:CMeter.hex:a
//avrdude -p m48p -c usbasp -U lfuse:w:0x77:m


#define F_CPU 16000000.0

#include <avr/io.h>
#include <avr/power.h>
#include <util/delay.h>
#include <avr/interrupt.h>


#define MEASURE_START INT0_vect
#define MEASURE_STOP  INT1_vect


//port D
#define DISCHARGE_PIN PD5
#define SEGMENT_A  PD6
#define SEGMENT_F  PD7


//led segments
#define SEGMENT_B PB0
#define SEGMENT_G PB1
#define SEGMENT_C PB2
#define SEGMENT_DP PB3
#define SEGMENT_D PB4
#define SEGMENT_E PB5

//port C
#define ANALOG_INPUT PC0
#define DIGIT_1 PC2
#define DIGIT_2 PC3
#define DIGIT_3 PC4
#define DIGIT_4 PC1


//measurement states
#define DATA_NOT_READY  0
#define DATA_READY_PF   1
#define DATA_READY_NF   2
#define DATA_OVERFLOW   3

#define RANGE_PF DATA_READY_PF
#define RANGE_NF DATA_READY_NF

#define CHAR_A 10
#define CHAR_C 11
#define CHAR_E 12
#define CHAR_H 13
#define CHAR_L 14
#define CHAR_R 15


#define PORT_B_NUMBER_MASK (1 << SEGMENT_B) | (1 << SEGMENT_C) | (1 << SEGMENT_D) | (1 << SEGMENT_E) | (1 << SEGMENT_G)
#define PORT_D_NUMBER_MASK (1 << SEGMENT_A) | (1 << SEGMENT_F)


//port B digit part
const uint8_t digit_B[] ={
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C) | (1 << SEGMENT_D) | (1 << SEGMENT_E)),                   //0
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C)),                                                         //1
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_D) | (1 << SEGMENT_E) | (1 << SEGMENT_G)),                   //2
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C) | (1 << SEGMENT_D) | (1 << SEGMENT_G)),                   //3
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C) | (1 << SEGMENT_G)),                                      //4
                        ~((1 << SEGMENT_C) | (1 << SEGMENT_D) | (1 << SEGMENT_G)),                                      //5
                        ~((1 << SEGMENT_C) | (1 << SEGMENT_D) | (1 << SEGMENT_E) | (1 << SEGMENT_G)),                   //6
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C)),                                                         //7
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C) | (1 << SEGMENT_D) | (1 << SEGMENT_E) | (1 << SEGMENT_G)),//8
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C) | (1 << SEGMENT_D) | (1 << SEGMENT_G)),                   //9

                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C) | (1 << SEGMENT_E) | (1 << SEGMENT_G)),                   //A
                        ~((1 << SEGMENT_D) | (1 << SEGMENT_E)),                                                         //C
                        ~((1 << SEGMENT_D) | (1 << SEGMENT_E) | (1 << SEGMENT_G)),                                      //E
                        ~((1 << SEGMENT_B) | (1 << SEGMENT_C) | (1 << SEGMENT_E) | (1 << SEGMENT_G)),                   //H
                        ~((1 << SEGMENT_D) | (1 << SEGMENT_E)),                                                         //L
                        ~((1 << SEGMENT_E) | (1 << SEGMENT_G)),                                                         //R
                        };

//port D digit part
const uint8_t digit_D[] ={
                            ~((1 << SEGMENT_A) | (1 << SEGMENT_F)),  //0
                            ~(0),                                    //1
                            ~((1 << SEGMENT_A)),                     //2
                            ~((1 << SEGMENT_A)),                     //3
                            ~((1 << SEGMENT_F)),                     //4
                            ~((1 << SEGMENT_A) | (1 << SEGMENT_F)),  //5
                            ~((1 << SEGMENT_A) | (1 << SEGMENT_F)),  //6
                            ~((1 << SEGMENT_A)),                     //7
                            ~((1 << SEGMENT_A) | (1 << SEGMENT_F)),  //8
                            ~((1 << SEGMENT_A) | (1 << SEGMENT_F)),  //9

                            ~((1 << SEGMENT_A) | (1 << SEGMENT_F)),  //A
                            ~((1 << SEGMENT_A) | (1 << SEGMENT_F)),  //C
                            ~((1 << SEGMENT_A) | (1 << SEGMENT_F)),  //E
                            ~((1 << SEGMENT_F)),                     //H
                            ~((1 << SEGMENT_F)),                     //L
                            ~((1 << SEGMENT_F)),                     //R
                        };


volatile uint8_t dataReady;
volatile uint8_t currentRange;
volatile uint8_t resultRange;
volatile uint8_t timerPrescaler;
volatile uint16_t ticks;
uint8_t strValue[4];

//number of "empty" ticks in pF and nF range is gained automatically during calibration at start up
uint16_t calibrationConstants[ DATA_OVERFLOW ];


//start measurement.
//Is trigered by external comparator when voltage across capacitor reach ~0.6V
ISR( MEASURE_START )
{
    //just start timer1
    TCCR1B = timerPrescaler;
}


//stop measurement.
//Is trigered by external comparator when voltage across capactor reach ~0.6V + Vref
ISR( MEASURE_STOP )
{
    //disable counter ( set prescaler to 0 )
    TCCR1B = 0;

    //copy tick vaule
    ticks = TCNT1;

    //set data ready flag
    dataReady = currentRange;

    //discharge capacitor
    PORTD |= ( 1 << DISCHARGE_PIN );
}


//when timer1 overflows. Capacity is too large
ISR( TIMER1_OVF_vect )
{
    //stop timer
    TCCR1B = 0;

    //set data ready value
    dataReady = DATA_OVERFLOW;

    //turn on discharging transistor
    PORTD |= ( 1 << DISCHARGE_PIN );
}


//convert binary value into decimal values
void intToStr( uint16_t n, uint8_t *result )
{
     uint8_t d3, d2, d1, d0, q;

     d0 = n       & 0x1F;
     d1 = (n>>5)  &  0xF;
     d2 = (n>>9)  &  0xF;
     d3 = (n>>13) &  0x7;

     d0 = ( (d3 + d2 + d1) << 1) + d0;
     q = (d0 * (uint16_t)0x67) >> 10;
     result[3] = d0 - 10*q;

     d1 = 9*d3 + d2 + 3*d1 + q;
     q = (d1 * (uint16_t)0x67) >> 10;
     result[2] = d1 - 10*q;

     d2 = d3 + 5*d2 + q;
     q = (d2 * (uint16_t)0x67) >> 10;
     result[1] = d2 - 10*q;

     d3 = (d3 << 3)+ q;
     q = (d3 * (uint16_t)0x1A) >> 8;
     result[0] = d3 - 10*q;
}


//set range
void setRange( uint8_t range )
{

    if( range == RANGE_PF )
    {
        //set rescaler to 1
        timerPrescaler = (1 << CS10 );

        //set a new range constant
        currentRange = RANGE_PF;
    }
    else
    {
        //set prescaler to 256
        timerPrescaler = ( 1 << CS12 );

        //set a new range constant
        currentRange = RANGE_NF;
    }
}


void turnOffDisplay()
{
    //turn off all driving (PNP) transistors
    PORTC |= ( 1 << DIGIT_1 ) | ( 1 << DIGIT_2 ) | ( 1 << DIGIT_3 ) | ( 1 << DIGIT_4 );
}


//set output pins according to number that will be displayed
void setNumber( uint8_t value )
{

    //first set all related PORT_B pins up.
    PORTB |= PORT_B_NUMBER_MASK;

    //now display character
    PORTB &= digit_B[ value ];

    //first set all related PORT_D pins up.
    PORTD |= PORT_D_NUMBER_MASK;

    //now display character
    PORTD &= digit_D[ value ];
}


//set digit on 4x7 segment display
void setDigit( uint8_t address )
{
    switch( address )
    {
        case 0:
            PORTC &= ~( 1 << DIGIT_1 );
            break;

        case 1:
            PORTC &= ~( 1 << DIGIT_2 );
            break;

        case 2:
            PORTC &= ~( 1 << DIGIT_3 );
            break;

        case 3:
            PORTC &= ~( 1 << DIGIT_4 );
            break;
    }
}


//display string from "strValue". Display is multiplexed and each "digit" is ON for about 1ms
//functions take almost constat time ~4ms to complete
void displayString()
{
    //turn off nF indicator dot
    PORTB |= ( 1 << SEGMENT_DP );

    //IF's are just for skipping initial 0s
    //digit 1
    if( strValue[0] != 0 )
    {
        turnOffDisplay();
        setNumber( strValue[0] );
        setDigit( 0 );
    }
    _delay_ms( 1 );


    //digit 2
    if( strValue[0] != 0 || strValue[1] != 0 )
    {
        turnOffDisplay();
        setNumber( strValue[1] );
        setDigit( 1 );
    }
    _delay_ms( 1 );


    //digit 3
    if( strValue[0] != 0 || strValue[1] != 0 || strValue[2] != 0 )
    {
        turnOffDisplay();
        setNumber( strValue[2] );
        setDigit( 2 );
    }
    _delay_ms( 1 );


    //digit 4
    turnOffDisplay();
    setNumber( strValue[3] );

    //add dot to the last digit if measured value is in nF range
    if( resultRange == RANGE_NF )
    {
        PORTB &= ~( 1 << SEGMENT_DP );
    }

    setDigit( 3 );
    _delay_ms( 1 );

    turnOffDisplay();
}


int main(void)
{

    //set clock prescaler to 1 to enable full 16MHz external clock
    clock_prescale_set( clock_div_1 );

    //set up int0 & int1
    //enable pullup resistors on PD3 and PD2 them ( are used by external open colector comparators )
    PORTD |= (1 << PD3 ) | ( 1 << PD2);

    //trigger int0 and int 1 on falling edge
    EICRA |= ( 1 << ISC01 ) | ( 1 << ISC11 );

    //enable int0 and int1
    EIMSK |= ( 1 << INT0 ) | ( 1 << INT1 );

    //set pins as output for addressing LED display and discharging capacitor
    DDRB |= ( 1 << SEGMENT_B ) | ( 1 << SEGMENT_C ) | ( 1 << SEGMENT_D ) | ( 1 << SEGMENT_E ) | ( 1 << SEGMENT_G ) | ( 1 << SEGMENT_DP );
    DDRD |= ( 1 << DISCHARGE_PIN ) | ( 1 << SEGMENT_A ) | ( 1 << SEGMENT_F );
    DDRC |= ( 1 << DIGIT_1 ) | ( 1 << DIGIT_2 ) | ( 1 << DIGIT_3 ) | ( 1 << DIGIT_4 );


    //timer1 is used to measure a time between int0 and int1
    //using (default) normal mode so there is nothing to do now.
    //just enable interupt for overflow (will flag a too large capacitor value)
    TIMSK1 |= ( 1 << TOIE1 );


    //enable global interupts
    sei();

    //discharge capacitor
    PORTD |= ( 1 << DISCHARGE_PIN );


    //calibrate both ranges.
    for( uint8_t range = RANGE_PF; range <= RANGE_NF; ++range )
    {
        setRange( range );

        //set CAL string
        strValue[0] = 0;
        strValue[1] = CHAR_C;
        strValue[2] = CHAR_A;
        strValue[3] = CHAR_L;

        //reset values
        dataReady = DATA_NOT_READY;
        TCNT1 = 0;
        resultRange = range;

        //display "CAL" string
        for( uint8_t i = 0; i < 100; ++i )
        {
            displayString();
        }

        //start measurement ( by turning off discharging transistor )
        PORTD &= ~( 1 << DISCHARGE_PIN );

        //wait until done
        while( dataReady == DATA_NOT_READY );

        //if timer overflows, there is too big capacity at input.
        if( dataReady == DATA_OVERFLOW )
        {
            //just stay in endless loop and display string "H"
            strValue[0] = 0;
            strValue[1] = 0;
            strValue[2] = 0;
            strValue[3] = CHAR_H;

            while( 1 )
            {
                displayString();
            }
        }

        //save the calibration value
        calibrationConstants[ range ] = ticks;

        //convert it to string and display it for a short time
        intToStr( ticks, strValue );

        //display calibration value
        for( uint8_t i = 0; i < 100; ++i )
        {
            displayString();
        }
    }


    //set range back to pF
    setRange( RANGE_PF );
    resultRange = RANGE_PF;


    //set defaul values
    strValue[0] = 0;
    strValue[1] = 0;
    strValue[2] = 0;
    strValue[3] = 0;
    TCNT1 = 0;
    dataReady = DATA_NOT_READY;

    //main loop
    while( 1 )
    {

        //if data are not ready ( we need to wait for them or start a new measurement )
        if( dataReady == DATA_NOT_READY )
        {
            //turn on measurement by switching off reseting/discharging transistor (if is already turned off nothing happen)
            PORTD &= ~( 1 << DISCHARGE_PIN );
        }

        //in the case of overflow ( too large capacitor )
        else if( dataReady == DATA_OVERFLOW )
        {
            //reset data flag and counter
            dataReady = DATA_NOT_READY;
            TCNT1 = 0;

            //and switch to higher range if possible
            if( currentRange == RANGE_PF )
            {
                setRange( RANGE_NF );
            }

            //if we reach the limit just display "H" and switch back to pF range
            else
            {
                strValue[0] = 0;
                strValue[1] = 0;
                strValue[2] = 0;
                strValue[3] = CHAR_H;

                setRange( RANGE_PF );
            }
        }

        //valid data in pF range
        else if( dataReady == DATA_READY_PF )
        {
            //reset data flag and counter (start a new measurement in the next iteration)
            dataReady = DATA_NOT_READY;
            TCNT1 = 0;

            //get number of ticks and substract calibration constant
            int32_t capacity = ( (int32_t)ticks - (int32_t)calibrationConstants[ RANGE_PF ] );


            //if value is negative there is wrong calibration or some drift over time/temperature, error or noise
            if( capacity < 0 )
            {
                //display "L"
                strValue[0] = 0;
                strValue[1] = 0;
                strValue[2] = 0;
                strValue[3] = CHAR_L;

                resultRange = RANGE_PF;
            }

            //value <= 9999pF can be displayed
            else if( capacity < 16000 )
            {
                //set the flag for pF range
                resultRange = RANGE_PF;

                //calculate the final capacity value from normalized ticks.
                capacity *= 10;
                capacity /= 16;

                //make string from integer that will be displayed
                intToStr( capacity, strValue );
            }

            //otherwise switch to nF range
            else
            {
                setRange( RANGE_NF );
            }
        }

        //valid data in nF range
        else if( dataReady == DATA_READY_NF )
        {
            //reset data flag and counter (start a new measurement in the next iteration)
            dataReady = DATA_NOT_READY;
            TCNT1 = 0;
            resultRange = RANGE_NF;

            //get number of ticks and remove initial offsets
            int32_t capacity = ( (int32_t)ticks - (int32_t)calibrationConstants[ RANGE_NF ] );

            //if value is negative there is wrong calibration or some drift over time/temperature, error or noise
            if( capacity < 0 )
            {
                //display "L"
                strValue[0] = 0;
                strValue[1] = 0;
                strValue[2] = 0;
                strValue[3] = CHAR_L;
            }

            //check if is lower then maximum displayable value
            else if( capacity < 62494 )
            {
                //calculate final value
                capacity *= 100;
                capacity /= 625;

                //make string from integer that will be displayed
                intToStr( capacity, strValue );
            }

            //otherwise capacity value is too large, Just print "H"
            else
            {
                strValue[0] = 0;
                strValue[1] = 0;
                strValue[2] = 0;
                strValue[3] = CHAR_H;
            }

            //switch back to pF range to start autoranging from beginning
            setRange( RANGE_PF );
        }

        //otherwise something strange happend. Just display Err
        else
        {
            strValue[0] = 0;
            strValue[1] = CHAR_E;
            strValue[2] = CHAR_R;
            strValue[3] = CHAR_R;
        }


        //display value for ~80*4 ms
        for( uint8_t i = 0; i < 80; ++i )
        {
            displayString();
        }
    }


    return 0;
}
